/*
 * Ben NanoNote LCD initialisation, based on uboot-xburst and xburst-tools.
 *
 * Copyright (C) 2001-2002 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef CONFIG_CPU_JZ4730_MINIPC
#include "minipc_claa070vc01.h"
#include "minipc.h"
#else
#include "nanonote_gpm940b0.h"
#include "nanonote.h"
#endif

#include "xburst_types.h"
#include "jzlcd.h"
#include "sdram.h"
#include "board.h"

extern vidinfo_t panel_info;
static unsigned long lcd_base;

static unsigned int get_line_length()
{
	return ALIGN((panel_info.vl_col * NBITS(panel_info.vl_bpix)) / 8, sizeof(u32));
}

static u32 *get_pixel32(unsigned short h, unsigned short v)
{
	return (u32 *) (lcd_base + v * get_line_length()) + h;
}

static u16 *get_pixel16(unsigned short h, unsigned short v)
{
	return (u16 *) (lcd_base + v * get_line_length()) + h;
}

static u8 *get_pixel8(unsigned short h, unsigned short v)
{
	return (u8 *) (lcd_base + v * get_line_length()) + h;
}

static u8 *get_pixel4(unsigned short h, unsigned short v)
{
	return (u8 *) (lcd_base + v * get_line_length()) + h / 2;
}

static inline unsigned short div(unsigned short num, unsigned short denom, unsigned short scale)
{
	return (scale * num) / denom;
}

static unsigned long pixel(unsigned short h, unsigned short v, u8 rmax, u8 gmax, u8 bmax, u8 rshift, u8 gshift, u8 bshift)
{
	unsigned short v_max = panel_info.vl_row;
	unsigned short h_max = panel_info.vl_col;

	return (((rmax - div(h, h_max, rmax + 1) + rmax - div(v, v_max, rmax + 1)) / 2) << rshift) |
		((gmax - div(v, v_max, gmax + 1)) << gshift) |
		(((div(h, h_max, bmax + 1) + div(v, v_max, bmax + 1)) / 2) << bshift);
}

static void test_pixel32(unsigned short h, unsigned short v)
{
	u32 *pix = get_pixel32(h, v);
	*pix = (u32) pixel(h, v, 255, 255, 255, 16, 8, 0);
}

static void test_pixel16_565(unsigned short h, unsigned short v)
{
	u16 *pix = get_pixel16(h, v);
	*pix = (u16) pixel(h, v, 31, 63, 31, 11, 5, 0);
}

static void test_pixel8(unsigned short h, unsigned short v)
{
	u8 *pix = get_pixel8(h, v);
	*pix = (u8) pixel(h, v, 7, 7, 3, 5, 2, 0);
}

static void test_pixel4(unsigned short h, unsigned short v)
{
	u8 *pix = get_pixel4(h, v);
	u8 mask = h & 1 ? 0xf0 : 0x0f;
	*pix = (*pix & mask) | ((u8) pixel(h, v, 1, 2, 1, 3, 1, 0) << (h & 1 ? 0 : 4));
}

void test_pixel(unsigned short h, unsigned short v)
{
	switch (panel_info.vl_bpix)
	{
		case LCD_COLOR32:
		test_pixel32(h, v);
		break;

		case LCD_COLOR8:
		test_pixel8(h, v);
		break;

		case LCD_COLOR4:
		test_pixel4(h, v);
		break;

		case LCD_COLOR16:
		default:
		test_pixel16_565(h, v);
		break;
	}
}

void clear_pixel32(unsigned short h, unsigned short v)
{
	u32 *pix = get_pixel32(h, v);
	*pix = 0;
}

void clear_pixel16(unsigned short h, unsigned short v)
{
	u16 *pix = get_pixel16(h, v);
	*pix = 0;
}

void clear_pixel8(unsigned short h, unsigned short v)
{
	u8 *pix = get_pixel8(h, v);
	*pix = 0;
}

void clear_pixel4(unsigned short h, unsigned short v)
{
	u8 *pix = get_pixel4(h, v);
	u8 mask = h & 1 ? 0xf0 : 0x0f;
	*pix = *pix & mask;
}

void clear_pixel(unsigned short h, unsigned short v)
{
	switch (panel_info.vl_bpix)
	{
		case LCD_COLOR32:
		clear_pixel32(h, v);
		break;

		case LCD_COLOR8:
		clear_pixel8(h, v);
		break;

		case LCD_COLOR4:
		clear_pixel4(h, v);
		break;

		case LCD_COLOR16:
		default:
		clear_pixel16(h, v);
		break;
	}
}

void test_pattern()
{
	unsigned short v_max  = panel_info.vl_row;
	unsigned short h_max  = panel_info.vl_col;
	unsigned short v, h;

	for (v = 0; v < v_max; v += 1) {
		for (h = 0; h < h_max; h += 1) {
			test_pixel(h, v);
		}
	}
}

void lcd_clear(unsigned long lcd_base)
{
	unsigned short v_max  = panel_info.vl_row;
	unsigned short h_max  = panel_info.vl_col;
	unsigned short v, h;
	unsigned long *pix = (unsigned long *)lcd_base;

	for (v = 0; v < v_max; v += 1) {
		for (h = 0; h < h_max; h += 1) {
			*pix++ = 0;
		}
	}
}

/* LCD initialisation. */

void lcd_init(void)
{
	__lcd_display_pin_init();
	__lcd_display_on();

	lcd_base = lcd_ctrl_init();
	lcd_clear(lcd_base);
	lcd_enable();
}
